package Forest::Dependency::Writer;

require 5.005_62;
use strict;
use warnings;
use Forest;
use XML::Generator;
use Forest::XML::Writer;

our $VERSION = '0.01';
our $forest;
our $dependency;
our %done;

our %color = ( 'subst' => 'blue',
	       'adj' => 'red',
	       'anchor' => 'yellow',
	       'lexical' => 'magenta',
	       'epsilon' => 'green'
	     );

our %trees;

our $dcounter;			# to assign id to derivations

our %derivations;

sub cluster {
    my $l = shift;
    my $r = shift;
    return "cluster$l"."_"."$r";
}

sub node {
  my $node = shift;
  my $label = cluster($node->{pos}[0],$node->{pos}[1])."_$node->{cat}_$node->{lemma}_".norm_tree(treename($node->{tree}));
  $label =~ s/\W//og;
  $label =~ s/_+/_/og;
  return $label;
}

sub treename {
  my $treename = shift;
  return (split(/\s+/,$treename))[0];
}

sub Forest::cluster_loop {
  my $self=shift;
  my $cluster_handler = shift;
   foreach my $posa (sort {$a <=> $b} keys %{$dependency}) {
     foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$posa}}) {
       $cluster_handler->($self,$posa,$posb,$dependency->{$posa}{$posb},@_);
     }
   }
}

sub Forest::node_loop {
  my ($self,$posa,$posb,$nodes,$node_handler,@l)=@_;
  foreach my $cat (keys %{$nodes->{tree}}) {
    foreach my $lemma (keys %{$nodes->{tree}{$cat}}) {
      foreach my $tree (keys %{$nodes->{tree}{$cat}{$lemma}}) {
	my $source = { pos => [$posa,$posb], 
		       cat => $cat, 
		       lemma => $lemma, 
		       tree => $tree, 
		       deriv => $nodes->{tree}{$cat}{$lemma}{$tree}{deriv} 
		     };
# 	print "NODE $posa $posb $cat $lemma $tree: ", 
# 	  join(" ", keys %{$nodes->{tree}{$cat}{$lemma}{$tree}{deriv}}),
# 	    "\n";
	$node_handler->($self,$source,$nodes->{tree}{$cat}{$lemma}{$tree},@l);
      }
    }
  }
}

sub Forest::edge_loop {
  my ($self,$source,$edges,$edge_handler,@l) = @_;
  foreach my $label (keys %{$edges->{label}}) {
    foreach my $target (@{$edges->{label}{$label}}) {
      $edge_handler->($self,$source,$target,$label,@l);
    }
  }
}

sub Forest::dependency_normalize {
    my $self=shift;
    my $nb = $self->{'nb'};
    $forest = $self;
    $dependency = {};
    %done=();
    %trees=();

    # First pass
    foreach my $head (keys %{$self->{'op'}}) {
	next if (defined $done{$head});
	$self->{'op'}{$head}->dependency();
    }

    # Second pass: emiting
    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
	foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	    my $lex = $dependency->{$pos}{$posb}{'lex'};
	    my @trees = keys %{$dependency->{$pos}{$posb}{'tree'}};
	    foreach my $tree (sort @trees) {
		my $cat = $trees{$tree}{'cat'};
		my $node = $trees{$tree}{'node'};
		my $edges = $dependency->{$pos}{$posb}{'tree'}{$tree};
		my $treename = treename($tree);
		foreach my $label (sort keys %{$edges}) {
		    foreach my $target (@{$edges->{$label}}) {
			my $tpos = $target->{'pos'}[0];
			my $tposb = $target->{'pos'}[1];
			my $ttree = $target->{'tree'};
			my $ttreename = treename($ttree);
			my $tlex = $dependency->{$tpos}{$tposb}{'lex'};
			my $tcat = $trees{$ttree}{'cat'};
			my $tnode = $trees{$ttree}{'node'};
			print "$target->{'type'}\t{$pos $posb $lex $treename $node $cat}\t{$tpos $tposb $tlex $ttreename $tnode $tcat}\t$label\n";
		    }
		}
	    }
	}
    }
}

sub tree_emit {
  my ($node,$grammar) = @_;
  my $treename = treename($node->{tree});
  my $xnode = node($node);
  my $cat = $node->{cat};
  my $url = (defined $grammar && defined $treename) ? ", URL=\"$grammar?id=$treename\"" : "";
  my $lemma = $node->{lemma} || '_';
  $treename ||= "";
  my $deriv = keys %{$node->{deriv} || {}};
  my $label = "$lemma:$cat:$treename ($deriv)";
  $label =~ s/"/\"/og;
  return <<EOF;
$xnode [label="$label", $url]
EOF
}

sub Forest::collect_dependency {
  my $self = shift;
  $dependency = {};
  $forest = $self;
  %done=();
  %trees=();
  %derivations=();

  $dcounter="d0000";

  foreach my $head (keys %{$self->{'op'}}) {
    next if (defined $done{$head});
    $self->{'op'}{$head}->dependency();
  }

}

sub Forest::xmldep {
    my $self=shift;

    # First pass: collecting dependencies
    $self->collect_dependency;

    my %nodes = ();
    my $nid= "n001";
    my $eid = "e001";
    my $GEN= new XML::Generator('encoding' => 'ISO-8859-1', 
				':pretty' => 1 );

    $Forest::XML::Writer::GEN=$GEN;

    my @clusters = ();
    my @nodes = ();
    my @edges = ();
    my @op = ();

    my $xmlnode = sub {
      my ($source,$flag) = @_;
      my $posa = $source->{pos}[0];
      my $posb = $source->{pos}[1];
      my $cat = $source->{cat};
      my $lemma = $source->{lemma};
      my $tree = $source->{tree};

#       print "NODE2 $posa $posb $cat $lemma $tree: ", 
# 	join(" ", keys %{$source->{deriv}}),
# 	  "\n";
      
      unless (exists $nodes{$posa}{$posb}{$cat}{$lemma}{$tree}) {
	$nodes{$posa}{$posb}{$cat}{$lemma}{$tree}{id} = $nid;
	++$nid;
      }

      if ($flag) {
	## The node should only be done for source nodes
	## target nodes are not fully equivalent to source nodes on their derivs
	## Sources nodes collect all derivs 
	## Target nodes collect only derivs for the edge that has been followed
	my $cluster = "c_".$posa."_".$posb;
	my %attr = ( id =>  $nodes{$posa}{$posb}{$cat}{$lemma}{$tree}{id},
		     cluster => $cluster,
		     tree => $tree || 'void',
		     cat => $cat,
		     lemma => $lemma
		   );
	$attr{deriv} = join(' ',keys %{$source->{deriv}}) if (keys %{$source->{deriv}});
	$attr{xcat} = $trees{$tree}{cat} if ($tree);
	push(@nodes,$GEN->node(\%attr));
      }

      return $nodes{$posa}{$posb}{$cat}{$lemma}{$tree}{id}
    };

    # Second pass: build xml
    my $edge_handler = sub { my ($self,$source,$target,$label,$ida) = @_;
			     my $idb = $xmlnode->( $target, 0 );
			     push(@edges,$GEN->edge({target =>$ida,
						     source =>$idb,
						     id =>$eid++,
						     type => $target->{type},
						     label => $label,
##						     deriv => join(' ',keys %{$target->{deriv}})
						    },
						    map($GEN->deriv({name => $_, 
								     span => join(' ',@{$target->{deriv}{$_}})}),
							keys %{$target->{deriv}})
						   ));
			   };
    my $node_handler = sub { my ($self,$source,$edges) = @_;
			     my $nid = $xmlnode->($source, 1);
			     $self->edge_loop( $source, $edges, $edge_handler, $nid );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				push(@clusters, $GEN->cluster({ id=> "c_".$posa."_".$posb,
								left => $posa,
								right => $posb,
								lex =>  $nodes->{'lex'} || ""
							      }));
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
			      };
    
    $self->cluster_loop( $cluster_handler );

    @op = $self->xmldep_op($GEN);

    # Third pass: emit
    print $GEN->xmldecl;
    my $attrs = {};
    $attrs->{grammar} = $self->{grammar} if (defined $self->{grammar});
    $attrs->{sentence} = $self->{sentence} if (defined $self->{sentence});
    $attrs->{mode} = $self->{mode} if (defined $self->{mode});
    print $GEN->dependencies($attrs,
			     @clusters,
			     @nodes,
			     @edges,
			     @op
			    ), "\n";

}


sub lp_quote {
  my $string = shift || "";
  return $string if ($string =~ /^[a-z_]\w*$/o);
  return $string if ($string =~ /^\d+$/);
  #    $string =~ s/\'/\'\'/og;
  $string =~ s/\'(?!\')/\'\'/og;
  return "\'$string\'";
}

sub Forest::lpdep {
    my $self=shift;

    # First pass: collecting dependencies
    $self->collect_dependency;

    my %nodes = ();
    my $nid= "n001";
    my $eid = "e001";

    my $lpnode = sub {
      my ($source) = @_;
      my $posa = $source->{pos}[0];
      my $posb = $source->{pos}[1];
      my $cat = $source->{cat};
      my $lemma = $source->{lemma};
      my $tree = $source->{tree};
      unless (exists $nodes{$posa}{$posb}{$cat}{$tree}) {
	my $cluster = "c_".$posa."_".$posb;
	my $qtree = lp_quote(treename($tree || 'void'));
	my $qcat = lp_quote($cat);
	my $qlemma = lp_quote($lemma);
	print "node{ id => $nid, cluster => $cluster, tree => $qtree, cat => $qcat, lemma => $qlemma }.\n"; 
	$nodes{$posa}{$posb}{$cat}{$tree}{id} = $nid;
	++$nid;
      }
      return $nodes{$posa}{$posb}{$cat}{$tree}{id}
    };
    
    print "%% sentence: $self->{sentence}\n" if (exists $self->{sentence});
    print "%% parser: $self->{parser}\n" if (exists $self->{parser});

    # Second pass: emit
    my $edge_handler = sub { my ($self,$source,$target,$label,$ida) = @_;
			     my $idb = $lpnode->( $target );
			     my $qlabel = lp_quote($label);
			     my $ide = $eid++;
			     print "edge{ target => $ida, source => $idb, id => $ide, type => $target->{type}, label => $qlabel }.\n";
			   };
    my $node_handler = sub { my ($self,$source,$edges) = @_;
			     $self->edge_loop( $source, $edges, $edge_handler, $lpnode->($source) );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				my $qlex = lp_quote($nodes->{'lex'});
				my $cluster = "c_".$posa."_".$posb;
				print "cluster{ id => $cluster, left => $posa, right => $posb, lex => $qlex }.\n";
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
			      };
    
    $self->cluster_loop( $cluster_handler );

}


sub Forest::stats {
    my $self=shift;
    
    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: compute stats
    my $sclusters = 0;
    my $snodes = 0;
    my $sedges = 0;
    
    my $edge_handler = sub { $sedges++; };
    my $node_handler = sub { my ($self,$source,$edges) = @_;
			     $snodes++;
			     $self->edge_loop( $source, $edges, $edge_handler );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				$sclusters++;
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
			      }; 
    
    $self->cluster_loop( $cluster_handler );
    
    printf "Dependency stats: %2.1f ambiguity %3d clusters %3d nodes %3d edges\n",
      ($sedges - $sclusters) / ($sclusters || 1),
	$sclusters,
	  $snodes,
	    $sedges;

}

sub Forest::dependency {
    my $self=shift;
    my $nb = $self->{'nb'};
    my $grammar = $self->{'grammar'};
    $forest = $self;

    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: emiting
    if (defined $grammar) {
      print "/* Grammar=$grammar */\n";
    }

    print <<HEADER;
digraph G {
    rankdir = LR;
    ranksep = .5;
    ratio = auto;
    clusterrank= local;
    node [  style = "filled",  fillcolor = "lightsalmon1"  ] ;
HEADER

    my $lex;
    my $cluster;
    my @trees;
    my $start;


    my $edge_handler = sub { my ($self,$source,$target,$label) = @_ ;
			     my $end = node($target);
			     my $color = $color{$target->{type}};
			     my $weight = abs($target->{pos}[0]-$source->{pos}[0]);
			     $weight = 1 unless ($weight);
			     my $deriv = keys %{$target->{deriv}};
			     if ($target->{pos}[0] > $source->{pos}[0]) {
			       print <<EDGE;
     $start -> $end [ label = "$label ($deriv)", color = "$color", minlen = $weight, dir = "back" ];
EDGE
			     } else {
			       print <<EDGE;
     $end -> $start [ label = "$label ($deriv)", color = "$color",  minlen = $weight ];
EDGE
			     }
			   };
    my $node_handler = sub { my ($self,$source,$edges) = @_;
			     push(@trees,tree_emit($source,$grammar));
			     $start = node($source);
			     $self->edge_loop( $source, $edges, $edge_handler );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				$cluster = cluster($posa,$posb);
				@trees = ();
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
				## Emit clusters
				my $lex = $nodes->{'lex'};
				my $trees = join(";\t\n",@trees);
				my $qlex = $lex;
				$qlex = "\\$qlex" if ($qlex eq '"');
				print <<ANCHOR;
subgraph  $cluster {
	    fontcolor = "magenta";
	    label = "$qlex";
            style = "filled";
            fillcolor = "lightyellow";
	    {rank = same; $trees; }
	}
ANCHOR
			      };

    $self->cluster_loop( $cluster_handler );

    print "}\n\n";
    
  }

sub Forest::Op::dependency {
    # return a list of anchors
    my $self=shift;
    my $tree= shift;
    my $head = $self->{'id'};
    return $done{$head} if ($done{$head});
    my $type = $self->{type};
    if ($type eq 'anchor' || $type eq 'lexical' || $type eq 'epsilon') {
      return unless (defined $tree);
      my $cat = $self->{cat};
      my $lemma = $self->{lemma} || "";
      my ($left,$right) = @{$self->{'span'}};
      my $entry =  $dependency->{$left}{$right} ||= {};
      $entry->{lex} ||= $self->{lex};
##      print "\nWARNING $type TREE ANCHOR $self->{tree} vs DERIV $tree\n";
##      $entry->{tree}{$cat}{$self->{tree}} ||= {};
      my $xtree = ($type eq 'lexical' || $type eq 'epsilon') ? "" : $tree;
      $entry->{tree}{$cat}{$lemma}{$xtree} ||= {};
      return $done{$head} = [ $self ];
    } else {
      my @anchors = ();		# the anchors of the various derivations for op
      foreach (@{ $self->{'derivations'}}) {
	push(@anchors,$_->dependency($head));
      }
      return $done{$head} = [ @anchors ];
    }
}

sub Forest::Deriv::dependency {
    # return an anchor
    my $self=shift;
    my $op=shift;
    my @anchors;
    my @edges = ();
    my $tree = $self->{tree};
    my $did = $self->{id} ||= $dcounter++;
    $trees{$tree} = { cat => $forest->{op}{$op}->{cat} , 
		      lemma => $forest->{op}{$op}->{lemma} || ""
		    };
    foreach my $label (keys %{$self->{'nodes'}}){
	foreach my $node (@{$self->{'nodes'}{$label}}){
	    next unless (defined $forest->{'op'}{$node});
	    my $target = $forest->{'op'}{$node};
	    if ($target->{'type'} eq 'anchor' && $target->{'span'}[0] < $target->{'span'}[1] ){
	      push(@anchors,$target);
	      $trees{$tree}{node} = $label;
	      next ;
	    }
	    if ($label =~ /^\d+/) {
		my $cat =  $target->{'cat'};
		$label = "$cat\#$label";
	    }
	    push(@edges, { 'type' => $target->{'type'},
			   'label' => $label,
			   'anchors' => $target->dependency($tree),
			   'targetspan' => $target->{span}
			 } 
		);
	}
    }
    ## WARNING: we may have cases of derivation with no anchor and no indication of tree
    ## what should we do ? maybe use the anchor of some node in deriv as anchor
    if (!@anchors) {
      ## We create a pseudo anchor
      my $xop = $forest->{'op'}{$op};
      my $dummy = Forest::Op->new( id => "$tree @{$xop->{span}}",
				   type => 'anchor',
				   lex => "",
				   lemma => "",
				   deriv => {},
				   tree => $tree,
##				   span => [$xop->{span}[0],$xop->{span}[1]],
				   span => [$xop->{span}[0],$xop->{span}[0]],
				   cat => $xop->{cat} 
				 );
      $dummy->dependency($tree);
      push(@anchors,$dummy);
    }
    my $anchor = $anchors[0];
    $anchor->dependency($tree);
    my $sanchor = $dependency->{$anchor->{span}[0]}{$anchor->{span}[1]}{tree}{$anchor->{cat}}{$anchor->{lemma} || ""}{$tree};

    # Skeletons of all possible derivation indexes
    # when ambiguity for the head of the target of a dependency, we have to duplicate the derivation id
    my @deriv_skel = ();
    foreach my $edge (@edges) {
      my @anchors = @{$edge->{anchors}};
      if (@anchors > 1 ) {
	my $i = 1;
	push(@deriv_skel,{ map( {$_ => $i++}  @anchors) });
      }
    }

    my @dids = map("$did$_",gen_did(0,@deriv_skel));
    $sanchor->{deriv}{$_} = 1 foreach (@dids);
    $self->{dids} = [@dids];

    foreach my $edge (@edges) {
	foreach my $start (@{$edge->{'anchors'}}) {
	  ## next unless $start;
	    $start->dependency();
	    my $sdep = $dependency->{$start->{'span'}[0]}{$start->{'span'}[1]}{'tree'};
	    $sdep = $sdep->{$start->{cat}}{$start->{lemma} || ""}{$start->{tree} || ""};
	    #	    print "/* add edge $edge->{'type'} from $start->{'tree'} $start->{'span'}[0] to  $anchor->{'tree'} $anchor->{'span'}[0] with label $edge->{'label'} */\n";
	    $sdep->{label}{$edge->{'label'}} ||= [];
	    my $info = { type => $edge->{'type'},
			 pos => $anchor->{'span'},
			 tree => $tree,
			 cat => $anchor->{cat},
			 lemma => $anchor->{lemma} || "",
			 deriv => { map( ("$did$_" => $edge->{targetspan}), gen_did($start,@deriv_skel)) }
		       };
	    next if (grep(&is_edge($_,$info),@{$sdep->{label}{$edge->{'label'}}}));
	    push(@{$sdep->{label}{$edge->{'label'}}},$info);
	}
    }

#     print "ANCHOR info: $anchor->{span}[0] $anchor->{span}[1] $anchor->{cat} $anchor->{lemma} $tree: ",
#       join(" ",keys %{$sanchor->{deriv}}),"\n";

    return $anchor;
}

sub gen_did {
  my $anchor = shift;
  my @skel = @_;
  return "" unless (@skel);
  my $anchors = shift(@skel);
  my @tail = gen_did($anchor,@skel);
  if (defined $anchors->{$anchor}) {
    return map("_$anchors->{$anchor}$_",@tail);
  } else {
    my @tmp = ();
    foreach my $v (values %{$anchors}) {
      push(@tmp,map("_$v$_",@tail));
    }
    return @tmp;
  }
}

sub is_edge {
    my $e1 = shift;
    my $e2 = shift;
    if ($e1->{'type'} eq $e2->{'type'} 
	&& $e1->{'pos'}[0] == $e2->{'pos'}[0] 
	&& $e1->{'pos'}[1] == $e2->{'pos'}[1] 
	&& $e1->{'cat'} eq $e2->{'cat'}
	&& $e1->{'lemma'} eq $e2->{'lemma'}
	&& $e1->{'tree'} eq $e2->{'tree'}) {
      ## Complete set of derivation using e1 with those using e2
      $e1->{deriv}{$_} = $e2->{deriv}{$_} foreach (keys %{$e2->{deriv}});
      return 1;
    } else {
      return 0;
    }
}

sub norm_tree {
    my $tree = shift || "";
#    print "/* norm tree $tree */\n";
    $tree =~ tr/-/_/;
    $tree =~ s/\s+/_/og;
    return $tree;
}


sub Forest::xmldep_op {
  my $self = shift;
  my $GEN = shift;

  my @op = ();

  foreach my $head (keys %{$self->{'op'}}) {
    push(@op,$self->{'op'}{$head}->xmldep_op($GEN));
  }
  
  return @op;
  
}

sub Forest::Op::xmldep_op {
  my $self = shift;
  my $GEN = shift;
  my $id = $self->{'id'};
  my $type = $self->{type};
  return if ($type eq 'anchor' || $type eq 'lexical');
  my @deriv = ();
  foreach (@{ $self->{'derivations'}}) {
    push(@deriv,$_->xmldep_op($GEN,$id));
  }
  my @content;
  push(@content,$self->{'top'}->pretty_print(2,'top')) if (defined $self->{'top'});
  push(@content,$self->{'bot'}->pretty_print(2,'bot')) if (defined $self->{'bot'});
  return $GEN->op({id=>$id,span=>join(' ',@{$self->{span}}), deriv => join(' ',@deriv)},@content);
}

sub Forest::Deriv::xmldep_op {
  my $self = shift;
  return @{$self->{dids}};
}

1;

__END__

=head1 NAME

Forest - Perl Objects to represent TAG Derivation Forest

=head1 SYNOPSIS

  use Forest;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=head1 SEE ALSO

perl(1).

=cut
