#
# Copyright (C) 2000, 2003, 2004, 2005 Eric de la Clergerie
# Forest::LP::Parser is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# $Id: Parser.pm,v 1.4 2005/11/03 13:29:45 clerger Exp $
#

package Forest::LP::Parser;

require 5.005;
use strict;
use warnings;
use Forest;
use Parse::RecDescent;

our $VERSION = '0.01';

#our $pattern = qr/(\w+)\s*(?:{(.*?)})?\(\s*(\d+)\s*,\s*(\d+)\s*\)/o;
our $pattern = qr/(\w+)\s*(?:{(.*?)})?\(\s*(\S+)\s*,\s*(\S+)\s*\)/o;
our $forest;
our %vars;
our %shared=();

our $type = 'lp';

our $config = { 'lp' => { 'adj' => \&lp_adj_parse,
			  'std' => \&lp_std_parse,
			  'anchor' => \&anchor_parse,
			  'lexical' => \&lexical_parse,
			  'struct' => \&struct_parse,
			  'epsilon' => \&epsilon_parse,
			},
		'xrcg' => { 'adj' => \&xrcg_adj_parse,
			    'std' => \&xrcg_std_parse,
			    'anchor' => \&anchor_parse,
			    'lexical' => \&lexical_parse
			    }
	    };

#$::RD_TRACE=1;

$::RD_AUTOACTION = q{ $item[1] };

my $prodgrammar = q{
       prod: head '<--' body { [$item{head},$item{body}] }
       body: '(' body ')' { $item{body} }
           | derivation(s /\|/) { [map(@$_,@{$item[1]})] }
           |      {[]}
       derivation:  comp(s) { 
                         Forest::LP::Parser::norm_derivation($item[1]) 
                         }
       comp: '[' label']' nt { [ $item{label} => $item{nt} ] }
           |  nt { [ 'void' => $item{nt} ] }
           |  shared    {  [ 'shared' => $item{shared} ] }
       nt: /\d+/ 
       label: /[\w<>_\d!-]+/ 
       shared: key '{'  body '}' { 
                      my $toto =  $Forest::LP::Parser::shared{$item{key}} ||= Forest::Shared->new(key=>$item{key});
                      $toto->{shared} = $item{body};
                      $toto
                    }
             | key { $Forest::LP::Parser::shared{$item{key}} ||=  Forest::Shared->new(key=>$item{key}) }
       key: '#' /\d+/ {$item[2]}
       head: nt 
};

my $prodparser = new Parse::RecDescent ($prodgrammar);

# $::RD_HINT=1;

my $fsgrammar = q{
      args : arg(s? /,/) { Forest::Feature->new( f => { map(@$_,@{$item[1]}) }) }
      fs: type '{' args '}' { $item{args}{type} = $item{type}; $item{args} }
      arg: feature '=>' value { $item{value}{comp} = $item{feature}; 
#                               print "Feature $item{feature}\n";
                                [$item{feature} => $item{value} ] }
      type: /[\w\d-]+/
      feature: /[\w\d-]+/
      value: var '::' rvalue { Forest::Comp->new( var => $item{var},  
                                                  value => $item{rvalue} ) }
           | var             { Forest::Comp->new( id => $item{var}, value => [] ) }
           | rvalue          { Forest::Comp->new( value => $item{rvalue} ) }
      rvalue: fs              { [ $item{fs} ]}
           | type '[' atomic(s /,/) ']' { $item[3] }
           | atomic          { [ $item{atomic} ] }
      atomic: /[a-z\d][\w\d+-]*/ { Forest::Val->new( val => $item[1]) }
           | '+'             { Forest::Plus->new() }
           | '-'             { Forest::Minus->new() }
      var: /[A-Z_][\w\d]*/
};

my $fsparser = new Parse::RecDescent ($fsgrammar);

sub lp_adj_parse {
    my $nt = shift;
    if ($nt =~ /^$pattern\s*\*\s*$pattern/o) {
	return { 'cat' => $1, 'top' => $2, 'bot' => $6, 'span' => [$3,$4,$7,$8] };
    } else {
	return 0;
    }
}

sub lp_std_parse {
    my $nt = shift;
    if ($nt =~ /^$pattern/o) {
	return { 'cat' => $1, 'top' => $2, 'span' => [$3,$4] };
    } else  {
	return 0;
    }
  }

sub struct_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!struct\(
                        \s*(.+?)\s*   # Tree
		 \)/ox) {
      return { 'tree' => $1 };
   } else {
     return 0;
   }
  }

sub anchor_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!anchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(.+?)\s*,    # Tree
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*\[(.+?),(.+?)\]\s*     # Lex
                 \)/ox) {
      return { 'lex' => $8, 'lemma' => $7, 'cat' => $5, 'bot' => $6, 'span' => [$2,$3], 'tree' => $4 };
    } elsif ($nt =~ m/^verbose!anchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(.+?)\s*,    # Tree
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*(.+?)\s*     # Lex
                 \)/ox) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $7, 'cat' => $5, 'bot' => $6, 'span' => [$2,$3], 'tree' => $4 };
    } elsif ($nt =~ /^tag_anchor\(\s*(.+?)\s*,\s*(\S+?)\s*,\s*(\S+?)\s*,\s*(.+?)\s*\)/o) {
      # Deprecated
      return { 'lex' => $1, 'span' => [$2,$3], 'tree' => $4 };
    } else {
      return 0;
    }
}

sub lexical_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!lexical\(
                         \s*\[\s*(.*?)\s*\]\s*,  ## Token
                         \s*(\S+?)\s*,            ## Left
                         \s*(\S+?)\s*,            ## Right
                         \s*(.+?)\s*,            ## Cat
                         \s*\[(.+?),(.+?)\]\s*             ## Lex
                 \)/xo) {
      return { 'lex' => $6, 'lemma' => $5, 'bot' => '', 'span' => [$2,$3], cat => $4 };
    } elsif ($nt =~ m/^verbose!lexical\(
                         \s*\[\s*(.*?)\s*\]\s*,  ## Token
                         \s*(\S+?)\s*,            ## Left
                         \s*(\S+?)\s*,            ## Right
                         \s*(.+?)\s*,            ## Cat
                         \s*(.+?)\s*             ## Lex
                 \)/xo) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $5, 'bot' => '', 'span' => [$2,$3], cat => $4 };
    }  elsif ($nt =~ m/^verbose!coanchor\(
		       \s*(.+?)\s*,    # Token
		       \s*(\S+?)\s*,    # Left
		       \s*(\S+?)\s*,    # Right
		       \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
		       \s*\[(.+?),(.+?)\]\s*     # Lex
                 \)/ox) {
      return { 'lex' => $7, 'lemma' => $6, 'cat' => $4, 'bot' => $5, 'span' => [$2,$3] };
    } elsif ($nt =~ m/^verbose!coanchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*(.+?)\s*     # Lex
                 \)/ox) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $6, 'cat' => $4, 'bot' => $5, 'span' => [$2,$3] };
    } elsif ($nt =~ /^tag_lexical\(\s*\[\s*(.*?)\s*\]\s*,\s*(\S+?)\s*,\s*(\S+?)\s*\)/o) {
      # Deprecated
      return { 'lex' => $1, 'span' => [$2,$3] };
    } else {
      return 0;
    }
}

sub epsilon_parse {
  my $nt = shift;
  if ($nt =~ m/^verbose!epsilon\(
	       \s*(.+?)\s*,             ## Token
	       \s*(\S+?)\s*,            ## Left
	       \s*(\S+?)\s*,            ## Right
	       \s*\[(.+?),(.+?)\]\s*    ## Lex
	       \)/xo) {
    ##    warn "FOUND EPSILON\n";
    return { 'lex' => $5, 'lemma' => $4, 'span' => [$2,$3] };
  } else {
    return 0;
  }
}

our $xrcg_pair = qr/\((\S+)\s*:\s*(\S+)\)/o;

sub xrcg_adj_parse {
    my $nt = shift;
    if ($nt =~ /^adj_(\w+)\(\w+(?:{(.*?)})?,\w+(?:{(.*?)})?\)\s*\($xrcg_pair,$xrcg_pair\)/o) {
	return { 'cat' => $1, 'top' => $2, 'bot' => $3, 'span' => [$4,$6,$5,$7] };
    } else {
	return 0;
    }
}

sub xrcg_std_parse {
    my $nt = shift;
    if ($nt =~ /^(\w+)(?:{(.*?)})?\s*\($xrcg_pair\)/o) {
	return { 'cat' => $1, 'top' => $2, 'span' => [$3,$4] };
    } else  {
	return 0;
    }
}

sub parse {
    my $handle = shift;
    $forest = Forest->new( {@_, 'op'=>{} } );
    %shared = ();

    $forest->{'nb'} = -1;

    while (<$handle>){
      if (/^\*+.*robust\s+mode/i) { # forest issued by partial robust parsing
	$forest->{mode} = 'robust';
	next;
      }
      last if (/^Shared Forest/);
    }

    while (<$handle>){
#	print "LINE $_";
	next if (/^\s*$/);	# skip blank lines
	if (/^\*ANSWER\*/) {	# skip ANSWER production
	    $_=<$handle>;
	    next;
	}
	s/__(\d+)/$1/og;		# Remove trailing number for variables
	my $nt=$_;
	my $production = <$handle>;
	chop $nt;
	chop $production;
	%vars=();
	my $info;
#    print "Dealing with $nt\n\t$production\n";
	my ($head,$derivations) = &handle_production($production);
	if ($info=&{$config->{$type}{'adj'}}($nt)) {
	    # Adjunctions
	    &var_set($info->{'top'});
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'adj',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'cat' => $info->{'cat'},
						      'top' => &handle_avm($info->{'top'}),
						      'bot' => &handle_avm($info->{'bot'})
						  );
	} elsif ($info=&{$config->{$type}{'std'}}($nt)) {
	    # Substitutions
	    &var_set($info->{'top'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'subst',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'cat' => $info->{'cat'},
						      'top' => &handle_avm($info->{'top'})
						      );
	} elsif ($info=&{$config->{$type}{'anchor'}}($nt)) {
	    # Anchors
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'anchor',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'lex' =>$info->{'lex'},
						      'bot' => &handle_avm($info->{'bot'}),
						      'cat' =>$info->{'cat'},
						      'tree' => $info->{'tree'},
						      lemma => $info->{lemma}
						      );
	} elsif ($info=&{$config->{$type}{'lexical'}}($nt)) {
	    # Lexicals
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'lexical',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'lex' =>$info->{'lex'},
						      'cat' =>$info->{'cat'},
						      'bot' => &handle_avm($info->{'bot'}),
						      lemma => $info->{lemma}
						    );
	  } elsif ($info=&{$config->{$type}{'struct'}}($nt)) {
	    # Structs
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'struct',
						      'derivations' => $derivations,
						      'tree' => $info->{'tree'}
						    );
	  } elsif ($nt =~ /^void_adj/) {
	    # deprecated old format
	  } elsif ($info=&{$config->{$type}{'epsilon'}}($nt)){
	    $forest->{'op'}{$head} = Forest::Op->new( id => $head,
						      type => 'epsilon',
						      derivations => $derivations,
						      span => $info->{'span'},
						      lex =>$info->{'lex'},
						      lemma => $info->{lemma}
						    );
	} else {
	    warn "Not a correct non terminal $nt";
	}
	$forest->{'nb'}=$head if ($head > $forest->{'nb'}) ;
    }
    ## A phrase of normalization
    ## To assign to each derivation its associated tree
    my @delete = ();
    foreach my $id (keys %{$forest->{'op'}}) {
      my $op = $forest->{'op'}{$id};
      my $type = $op->{type};
      if ($type eq 'struct') {
	push(@delete,$id);
	next;
      }
      foreach my $derivation (@{$op->{derivations}}) {
	foreach my $node (keys %{$derivation->{nodes}}) {
	  foreach my $subid (@{$derivation->{'nodes'}{$node}}) {
	    my $subop = $forest->{'op'}{$subid};
	    $derivation->{tree} = $subop->{tree} if (defined $subop->{tree});
	    ## Should remove struct op
	  }
	}
      }
    }
    delete $forest->{op}{$_} foreach (@delete);
    return $forest;
}

sub handle_production {
    my $production = shift;
    return @{$prodparser->prod($production)};
}

sub norm_derivation {
    my $deriv = shift;
    my @allderiv = ({});
    my @deriv = @{$deriv};
    while (@deriv) {
        my $lnt = shift @deriv;
	my ($label,$node) = @$lnt;
	if ($label eq 'shared') {
	  my @sderiv = map($_->{nodes},@{$node->{shared}});
	  my @tmpderiv = ();
	  foreach my $sderiv (@sderiv) {
	    foreach my $nderiv (@allderiv) {
	      my $tmpderiv = {};
	      foreach my $key (keys %$sderiv) {
		$tmpderiv->{$key} = [ @{$sderiv->{$key}} ];
	      }
	      foreach my $key (keys %$nderiv) {
		push(@{$tmpderiv->{$key}},@{$nderiv->{$key}});
	      }
	      push(@tmpderiv,$tmpderiv);
	    }
	  }
	  @allderiv = @tmpderiv;
	} else {
	  foreach my $nderiv (@allderiv) {
	    push(@{$nderiv->{$label}},$node);
	  }
	}
    }
    return [map(Forest::Deriv->new(nodes=>$_),@allderiv)];
}

sub handle_avm {
  my $avm = shift || '';
  return $fsparser->args($avm);
}

sub handle_value {
    shift;
    s/^\s*//o;
    s/\s*$//;
    return Forest::Plus->new() if /^\+$/;
    return Forest::Minus->new() if /^-$/;
    return Forest::Val->new( val => $_ );
}

sub var_set {
  my $p = shift;
  return unless $p;
  $vars{$_}++ foreach ($p =~ /[A-Z_]\w*/og);
}

package Forest::LP::XRCG::Parser;

sub parse {
    $Forest::LP::Parser::type='xrcg';
  my $forest = Forest::LP::Parser::parse(@_);
    $Forest::LP::Parser::type='lp';
    return $forest;
}

1;
__END__

=head1 NAME

Forest::LP::Parser - Parse the LP representation of TAG Derivation forest

=head1 SYNOPSIS

  use Forest::LP::Parser;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=head1 SEE ALSO

Forest (3)
perl(1).

=cut


sub handle_avm {
    my %fs = ();
    my $avm = shift || '';
    $avm =~ /^/g;
    while ($avm =~ /\G\s*(\w+)\s*=>/gc) {
	my $f = $1;
	my $var = "";
	my $val = "";
	my @trail = ();
	$var = $1 if ($avm =~ /\G\s*([A-Z]\w*)\s*(::)?/gc);
	$val = $1 if ($avm =~ /\G\s*([[:alnum:]+-]+)/gc);
	@trail = split(',',$1) if ($avm =~ /\G[\{\[](.*?)[\}\]]/gc);
	$avm =~ /\G\s*,/gc;
	$var="" if ($var && $vars{$var} == 1);
	@trail = ($val) if (!@trail && $val);
	@trail = map( &handle_value($_), @trail);
	my $comp = Forest::Comp->new( 'comp' => $f, 
				      'value' => [@trail] );
	$comp->{'id'} = $var if ($var);
	$fs{$f}=$comp;
    }
    return Forest::Feature->new( 'f' => \%fs );
}
