package Forest::XML::Parser;

require 5.005_62;
use strict;
use warnings;
use Forest;
use XML::Parser;

our $VERSION = '0.01';
our @ISA = qw{XML::Parser};

our $forest;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(Pkg=>__PACKAGE__,
                                  Style=>'Subs',
                                  @_);
    $self->setHandlers(Final => \&Final,
                       Init => \&Init,
                       Char => \&Char
                       );
    return $self;
}

sub parse {
  my $expat = Forest::XML::Parser->new();
  return $expat->SUPER::parse(@_);
}

sub Char {
  my $expat = shift;
  my $text = shift;
  $text =~ s/^\s+//o;
  $text =~ s/\s+$//o;
  $expat->{Current}{$expat->{'Where'}}=$text unless ($text  =~ /^\s*$/);
}

sub Final {
  my $expat = shift;
  delete $expat->{Current};
  delete $expat->{Where};
  delete $expat->{Stack};
  return $expat->{Forest};
}

sub Init {
    my $expat = shift;
    $expat->{Forest} = Forest->new( {'op'=> {}, 'nb' => 0} );
}

sub forest {
    my $expat = shift;
    my $tag = shift;
    my %attr = @_;
    $expat->{Forest}{'parser'} = $attr{'parser'} if (defined $attr{'parser'});
    $expat->push($expat->{Forest});
}

sub forest_ {
    shift->pop();
}

sub sentence {
  my $expat = shift;
    $expat->{'Where'} = 'sentence';
}

sub sentence_ {
    my $expat = shift;
    undef $expat->{Where};
}

sub op {
    my $expat = shift;
    my $tag = shift;
    my %attr = @_;
    my $op = Forest::Op->new('id'=>$attr{'id'},
			     'span'=>[ split(/\s+/,$attr{'span'}) ],
			     'cat' => $attr{'cat'},
			     'type' => $attr{'type'},
			     'derivations' => []);
    $expat->{'Forest'}{'op'}{$op->{'id'}}=$op;
    $expat->{'Forest'}{'nb'} = $op->{'id'} if ($op->{id} > $expat->{'Forest'}{'nb'});
    $expat->push($op);
}

sub op_ {
    shift->pop();
}

sub narg {
    my $expat = shift;
    shift;
    my %attr = @_;
    my $node = $expat->{Current};
    $expat->{Where} = $attr{'type'};
}

sub narg_ {
    my $expat = shift;
    undef $expat->{Where};
}

sub plus {
    my $expat = shift;
    my $plus = Forest::Plus->new();
    push( @{ $expat->{Current}{'value'} },$plus);
}

sub minus {
    my $expat = shift;
    my $minus = Forest::Minus->new();
    push( @{ $expat->{Current}{'value'} },$minus);
}

sub fs {
    my $expat = shift;
    my $feature = Forest::Feature->new();
    my $where = $expat->{Where} || 'feature';
    $expat->{Current}{$where} = $feature;
    $expat->push($feature);
}

sub fs_ {
    shift->pop();
}

sub f {
    my $expat = shift;
    my $tag = shift;
    my $comp = Forest::Comp->new( { @_, value =>[] } );
    $expat->{Current}{'f'}{$comp->{'name'}} = $comp;
    $expat->push($comp);
}

sub f_ {
    shift->pop;
}

sub val {
    my $expat = shift;
    my $tag = shift;
    my $val = Forest::Val->new();
    push( @{ $expat->{Current}{'value'} },$val);
    $expat->push($val);
}

sub val_ {
    shift->pop;
}

sub link {
    my $expat = shift;
    my $tag = shift;
    my %attr = @_;
    $expat->{Current}{'id'}=$attr{'id'};
}

sub deriv {
    my $expat = shift;
    my $tag = shift;
    my %attr = @_;
    my $anchor = $attr{'node'} || '<>';
    my $deriv = Forest::Deriv->new( 
				    'nodes' => { $anchor => 
						   Forest::Node->new(
								     )
						 } );
    push(@{$expat->{Current}{'derivations'}},$deriv);
    $expat->push($deriv);
}

sub deriv_ {
    shift->pop;
}


1;

__END__
